class Box:
    """ constants class for recommended images and boxes """
    UBUNTU_JAMMY64 = 'ubuntu/jammy64'
    UBUNTU_FOCAL64 = 'ubuntu/focal64'
    UBUNTU_BIONIC64 = 'ubuntu/bionic64'
    UBUNTU_XENIAL64 = 'ubuntu/xenial64'
    CLOUDIMAGE_JAMMY = 'https://cloud-images.ubuntu.com/jammy/current/jammy-server-cloudimg-amd64.img'
    CLOUDIMAGE_FOCAL = 'https://cloud-images.ubuntu.com/focal/current/focal-server-cloudimg-amd64.img'
    CLOUDIMAGE_BIONIC = 'https://cloud-images.ubuntu.com/bionic/current/bionic-server-cloudimg-amd64.img'
    CLOUDIMAGE_XENIAL = 'https://cloud-images.ubuntu.com/xenial/current/xenial-server-cloudimg-amd64-disk1.img'
    CLOUDIMAGE_JAMMY_ARM = 'https://cloud-images.ubuntu.com/jammy/20230330/jammy-server-cloudimg-arm64.img'
    DOCKER_JAMMY = 'ubuntu:jammy'
    DOCKER_FOCAL = 'ubuntu:focal'
    DOCKER_BIONIC = 'ubuntu:bionic'
    DOCKER_XENIAL = 'ubuntu:xenial'
